#include <GDD/ddg.h>
#include <SIRA/loop.h>
#include <SIRA/isolver.h>

int main(int argc, char *argv[]) {
  // two operations
  GDD_Opcode *i1 =  GDD_opcode_create("inst_1", 1, 0);
  GDD_opcode_extend_delta_w(i1, "GR", 0);

  GDD_Opcode *i2 =  GDD_opcode_create("inst_2", 1, 0);
  GDD_opcode_extend_delta_w(i2, "GR", 0);

  GDD_Opcode *i3 =  GDD_opcode_create("inst_3", 1, 0);
  GDD_opcode_extend_delta_w(i3, "GR", 0);

  GDD_Opcode *i4 =  GDD_opcode_create("inst_4", 1, 0);
  GDD_opcode_extend_delta_w(i4, "GR", 0);

  GDD_Opcode *i5 =  GDD_opcode_create("inst_5", 1, 0);
  GDD_opcode_extend_delta_w(i5, "GR", 0);

  GDD_Opcode *i6 =  GDD_opcode_create("inst_6", 1, 0);
  GDD_opcode_extend_delta_w(i6, "BR", 0);

  GDD_Opcode *i7 =  GDD_opcode_create("inst_7", 1, 0);

  // build the architecture description
  GDD_Arch *arch = GDD_arch_create();
  GDD_arch_add_regtype(arch, "GR", 61);
  GDD_arch_add_regtype(arch, "BR", 8);
  GDD_arch_add_opcode(arch, i1);
  GDD_arch_add_opcode(arch, i2);
  GDD_arch_add_opcode(arch, i3);
  GDD_arch_add_opcode(arch, i4);
  GDD_arch_add_opcode(arch, i5);
  GDD_arch_add_opcode(arch, i6);
  GDD_arch_add_opcode(arch, i7);

  // build a DDG
  SCEDA_Graph *g = SCEDA_graph_create((SCEDA_delete_fun)GDD_node_label_delete, (SCEDA_delete_fun)GDD_edge_label_delete); 

  SCEDA_Vertex *v1 = SCEDA_graph_add_vertex(g, GDD_node_label_create(1, i1));
  SCEDA_Vertex *v2 = SCEDA_graph_add_vertex(g, GDD_node_label_create(2, i2));
  SCEDA_Vertex *v3 = SCEDA_graph_add_vertex(g, GDD_node_label_create(3, i3));
  SCEDA_Vertex *v4 = SCEDA_graph_add_vertex(g, GDD_node_label_create(4, i4));
  SCEDA_Vertex *v5 = SCEDA_graph_add_vertex(g, GDD_node_label_create(5, i5));
  SCEDA_Vertex *v6 = SCEDA_graph_add_vertex(g, GDD_node_label_create(6, i6));
  SCEDA_Vertex *v7 = SCEDA_graph_add_vertex(g, GDD_node_label_create(7, i7));

  SCEDA_graph_add_edge(g, v1, v2, GDD_edge_label_create("GR", "flowdep_reg", 3, 0));
  SCEDA_graph_add_edge(g, v1, v7, GDD_edge_label_create(NULL, "serial", 0, 0));
  SCEDA_graph_add_edge(g, v2, v3, GDD_edge_label_create("GR", "flowdep_reg", 3, 0));
  SCEDA_graph_add_edge(g, v3, v3, GDD_edge_label_create("GR", "flowdep_reg", 1, 1));
  SCEDA_graph_add_edge(g, v3, v7, GDD_edge_label_create("GR", "flowdep_reg", 1, 0));      
  SCEDA_graph_add_edge(g, v4, v4, GDD_edge_label_create("GR", "flowdep_reg", 1, 1));
  SCEDA_graph_add_edge(g, v4, v6, GDD_edge_label_create("GR", "flowdep_reg", 1, 0));
  SCEDA_graph_add_edge(g, v5, v1, GDD_edge_label_create(NULL, "serial", 1, 4));
  SCEDA_graph_add_edge(g, v5, v5, GDD_edge_label_create("GR", "flowdep_reg", 1, 1));
  SCEDA_graph_add_edge(g, v6, v7, GDD_edge_label_create("BR", "flowdep_reg", 3, 0));

  // create a siraloop with MIIrec = 2
  SIRA_Loop *loop = SIRA_loop_create(g, 2);

  // initialise settings for the solver
  SIRA_Param settings;
  SIRA_param_init(&settings);
  SIRA_param_set_ual_flag(&settings, FALSE);
  SIRA_param_set_bounded_flag(&settings, FALSE);
  SIRA_param_set_check_flag(&settings, TRUE);

  // build the desired list of types
  SCEDA_List *types = SCEDA_list_create((SCEDA_delete_fun)free);
  SCEDA_list_add(types, strdup("GR"));

  SCEDA_HashMap *solution;
  SCEDA_HashMap *costmap;
  if(SIRA_isolver_single(loop, &settings, types, 2, &solution, &costmap) == 0) {
    fprintf(stdout,"siralina found a solution\n");

    // check whether the resuling DDG is lexicographic positive
    SCEDA_List *cycle;
    if(GDD_is_lexicographic_positive(SIRA_loop_ddg(loop), &cycle)) {
      fprintf(stderr,"resulting DDG is lexicographic positive\n");
    } else {
      fprintf(stderr,"resulting DDG is NOT lexicographic positive\n");
      SCEDA_list_delete(cycle);
    }

    // undo the modifications 
    // this is absolutely needed before deleting the loop and the DDG!
    SIRA_isolver_undo(loop, solution);

    // clear the solution and the cost map
    SCEDA_hashmap_delete(solution);
    SCEDA_hashmap_delete(costmap);
  }

  // clear the list of types
  SCEDA_list_delete(types);

  // clean the settings (free some internal data structures)
  SIRA_param_cleanup(&settings);

  SIRA_loop_delete(loop); // delete the loop

  SCEDA_graph_delete(g); // delete g (not that reuse decisions must be unapplied before this call!)
  GDD_arch_delete(arch);     // and the architecture

  return 0;
}

