#include <stdio.h>
#include <libxml/parser.h>

#include <GDD/arch_xml.h>
#include <GDD/ddg_xml.h>
#include <GDD/ddg.h>
#include <SIRA/sira.h>

#include <assert.h>

int main(int argc, char *argv[]) {
  // initialise XML parser
  xmlInitParser();

  // read architecture
  GDD_Arch *arch = GDD_arch_of_xml_file("arch.xml");
  assert(arch != NULL);

  // read ddg (ignore interloop dependencies)
  SCEDA_Graph *dag = GDD_of_xml_file(TRUE, arch, "ddg.xml", NULL);
  assert(dag != NULL);

  // read ddg 
  int mii_rec;
  SCEDA_Graph *ddg = GDD_of_xml_file(FALSE, arch, "ddg.xml",&mii_rec);
  assert(ddg != NULL);

  // clean up XML parser
  xmlCleanupParser();

  // reduction of acyclic saturation
  {
    fprintf(stdout,"reduction of acyclic saturation\n");
    // set weight of types
    SIRA_WeightMap *weights = SIRA_weight_map_create();
    SIRA_weight_map_set_weight_of_type(weights, "GR", 1); // reduce GR saturation
    SIRA_weight_map_set_weight_of_type(weights, "BR", 1); // reduce BR saturation

    SCEDA_List *new_edges;
    if(SIRA_try_reduce_acyclic_saturation(dag, weights, &new_edges) == 0) {
      fprintf(stdout,"%d edges added.\n",SCEDA_list_size(new_edges));
      GDD_delete_edges(dag, new_edges);
      SCEDA_list_delete(new_edges);
    } else {
      fprintf(stdout,"failed to reduce saturation\n");
    }

    SIRA_weight_map_delete(weights);
  }

  // reduction of cyclic saturation (conjointly)
  {
    fprintf(stdout,"reduction of cyclic saturation (conjointly)\n");
    // create the sira loop
    SIRA_Loop *loop = SIRA_loop_create(ddg, mii_rec);

    // set register constraints
    SIRA_RegConstraintsMap *constraints = SIRA_regconstraints_map_create();
    SIRA_regconstraints_map_set_constraint_of_type(constraints, "GR", GDD_arch_get_regtype(arch, "GR"));
    SIRA_regconstraints_map_set_constraint_of_type(constraints, "BR", GDD_arch_get_regtype(arch, "BR"));
    
    {
      // initialise settings for the solver
      SIRA_Param settings;
      SIRA_param_init(&settings);
      
      // set weight of types
      SIRA_WeightMap *weights = SIRA_weight_map_create();
      SIRA_weight_map_set_weight_of_type(weights, "GR", 1); // reduce GR saturation
      SIRA_weight_map_set_weight_of_type(weights, "BR", 1); // reduce BR saturation
      
      // try to reduce cyclic saturation
      SCEDA_List *new_edges;
      if(SIRA_try_reduce_cyclic_saturation(loop, &settings, constraints, weights, &new_edges) == 0) {
	fprintf(stdout,"%d edges added.\n",SCEDA_list_size(new_edges));
	GDD_delete_edges(SIRA_loop_ddg(loop), new_edges);
	SCEDA_list_delete(new_edges);
	SIRA_loop_update(loop);
      } else {
	fprintf(stdout,"failed to reduce saturation\n");
      }
      
      // delete weights
      SIRA_weight_map_delete(weights);
      
      // clean up settings
      SIRA_param_cleanup(&settings);
    }

    // reduction of cyclic saturation (iteratively)
    {
      fprintf(stdout,"reduction of cyclic saturation (iteratively)\n");
      // initialise settings for the solver
      SIRA_Param settings;
      SIRA_param_init(&settings);
      
      // initialise list of type
      SCEDA_List *types = SCEDA_list_create(free);
      SCEDA_list_add(types, strdup("GR"));
      SCEDA_list_add(types, strdup("BR"));
      
      SCEDA_List *new_edges;
      if(SIRA_try_reduce_cyclic_saturation_iteratively(loop, &settings, constraints, types, &new_edges) == 0) {
	fprintf(stdout,"%d edges added.\n",SCEDA_list_size(new_edges));
	GDD_delete_edges(SIRA_loop_ddg(loop), new_edges);
	SCEDA_list_delete(new_edges);
	SIRA_loop_update(loop);
      } else {
	fprintf(stdout,"failed to reduce saturation\n");
      }
      
      SCEDA_list_delete(types);
      
      // clean up settings
      SIRA_param_cleanup(&settings);
    }
    
    // delete constraints
    SIRA_regconstraints_map_delete(constraints);

    // delete the sira loop
    SIRA_loop_delete(loop);
  }

  // delete the graphs and the architecture
  SCEDA_graph_delete(dag);
  SCEDA_graph_delete(ddg);
  GDD_arch_delete(arch);

  return 0;
}
