/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
/** \file regconstraints.h
    \brief Setting register constraints */
#ifndef __SIRA_REGCONSTRAINTS_H
#define __SIRA_REGCONSTRAINTS_H

#include <SCEDA/hashmap.h>

/** Type of the register contraints map.

    \hideinitializer */
typedef SCEDA_HashMap SIRA_RegConstraintsMap;

/** Create an empty register constraints map.

    @return a new register constraints map.
*/
SIRA_RegConstraintsMap *SIRA_regconstraints_map_create();

/** Delete a register constraints map.

    @param[in] regconstraints_map = register constraints map to delete 

    \hideinitializer */
#define SIRA_regconstraints_map_delete SCEDA_hashmap_delete

/** Set register constraints of given type.

    @param[in] regconstraints_map = register constraints map
    @param[in] type = type
    @param[in] constraint = maximal number of register of given type

    @return 0 in case of success, <>0 otherwise */
int SIRA_regconstraints_map_set_constraint_of_type(SIRA_RegConstraintsMap *regconstraints_map, const char *type, int constraint);

/** Set register constraints of given type.

    @param[in] regconstraints_map = register constraints map
    @param[in] type = type

    @return constraint associated to given type or INT_MAX when undefined */
int SIRA_regconstraints_map_get_constraint_of_type(SIRA_RegConstraintsMap *regconstraints_map, const char *type);

/** Clear the register constraints map.

    @param[in] regconstraints_map = register constraints map to clear 

    \hideinitializer */
#define SIRA_regconstraints_map_clear SCEDA_hashmap_clear

#endif

/** \page regconstraints_map Setting register constraints

    Register constraints are defined in a SIRA_RegConstraintsMap structure.

    A SIRA_RegConstraintsMap is created thanks to SIRA_regconstraints_map_create
    function and deleted thanks to SIRA_regconstraints_map_delete function.
    
    We advise to access the SIRA_RegConstraintsMap structures
    exclusively through the functions provided by the API.

    \section regconstraints_map_api API

    \code
    SIRA_RegConstraintsMap *SIRA_regconstraints_map_create();
    \endcode

    Create an empty register constraints map.

    \code
    void SIRA_regconstraints_map_delete(SIRA_RegConstraintsMap *regconstraints_map);
    \endcode

    Delete a register constraints map.

    \code
    int SIRA_regconstraints_map_set_constraint_of_type(SIRA_RegConstraintsMap *regconstraints_map, const char *type, int constraint);
    \endcode

    Set maximal number of registers of given type in the register constraints map.

    \code
    int SIRA_regconstraints_map_get_constraint_of_type(SIRA_RegConstraintsMap *regconstraints_map, const char *type);
    \endcode

    Get maximal number of registers of given type in the register constraints map.

    \code
    void SIRA_regconstraints_map_clear(SIRA_RegConstraintsMap *regconstraints_map);
    \endcode

    Clear the given register constraints map.

*/
