/*
   Copyright Universite de Versailles Saint-Quentin en Yvelines 2009
   AUTHORS: Sebastien Briais, Sid-Ahmed-Ali Touati

   This file is part of SIRA.
   
   SIRA is free software: you can redistribute it and/or modify it
   under the terms of the GNU Lesser General Public License as
   published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.
   
   SIRA is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
   
   You should have received a copy of the GNU Lesser General Public
   License along with SIRA.  If not, see
   <http://www.gnu.org/licenses/>.
*/
#ifndef __SIRA_SOLVER_H
#define __SIRA_SOLVER_H
/** \file solver.h
    \brief SiraLina solver */

#include "loop.h"
#include "reuse.h"
#include "param.h"
#include "weightmap.h"
#include "regconstraints.h"

/** Apply siralina heuristic for every II between IImin and IImax and
    print the results (used for statistics).

    @param[in] loop = siraloop
    @param[in] settings = solver settings
    @param[in] weight_map = map giving weight of types to consider
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) */
void SIRA_solver_stats(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax);

/** Apply siralina heuristic for given II.

    @param[in] loop = siraloop
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] weight_map = map giving weight of types to consider
    @param[in] II = II (if negative, fails)
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return 0 if a solution was found, otherwise -1 */
int SIRA_solver_single(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int II, SCEDA_HashMap **solution);

/** Apply siralina heuristic for every II between IImin and IImax and
    stops when a solution is found (ie satisfies the register constraints).

    Before beginning the search procedure, a simple heuristic is
    implemented to determine whether there is a solution to the
    problem.

    @param[in] loop = siraloop
    @param[in] constraints = register contraints
    @param[in] settings = solver settings (semantics might be modified)
    @param[in] weight_map = map giving weight of types to consider
    @param[in] IImin = IImin (if negative, IImin of the loop will be used)
    @param[in] IImax = IImax (if negative, IImax of the loop will be used) 
    @param[out] solution = computed solution (map from types to reuse edges list)

    @return II >= 0 if a solution was found, otherwise -1 */
int SIRA_solver_search(SIRA_Loop *loop, SIRA_RegConstraintsMap *constraints, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax, SCEDA_HashMap **solution);

#endif

/** \page solver SiraLina solver 

    \section solver_intro Introduction

    The SiraLina solver implements siralina heuristic. Siralina is an
    approximate method that aims at solving the problem of periodic
    task scheduling with storage minimisation. We review briefly
    siralina method in the following.

    To attach more or less importance to a given register type, each
    type \f$t\f$ comes with a weight \f$\alpha_t\f$.

    SiraLina proceeds in two steps: first it solves the <em>scheduling
    problem</em>, and then it solves (several instances of) the
    <em>assignment problem</em>. Let \f$p\f$ be the desired period for
    the loop.
    
    We first review the definition of the scheduling problem.

    For each operation of the DDG, we define a schedule variable
    \f$\sigma_v\f$. Moreover, for each type and each value of the DDG
    of given type, we define a schedule variable \f$\sigma_{k_v}^t\f$.

    The scheduling problem is the following linear integer programming
    problem:
    
    \li minimise \f$\sum\limits_{t \in \mathcal{T}} \alpha_t
    \sum\limits_{v \in V_R^t} (\sigma_{k_v}^t - \sigma_v)\f$

    \li subject to \f$\sigma_v - \sigma_u \geq \delta(e) - p \times
    \lambda(e)\f$ for each edge \f$e \in E\f$ with
    \f$\mathop{source}(e) = u\f$ and \f$\mathop{target}(e) = v\f$

    \li and to \f$\sigma_{k_u}^t - \sigma_v \geq \delta_r^t(v) + p
    \times \lambda(e)\f$ for each type \f$t\f$ and flow dependency
    edge \f$e \in E_R^t\f$ with \f$\mathop{source}(e) = u\f$ and
    \f$\mathop{target}(e) = v\f$

    (if, for a given type \f$t\f$, value \f$u\f$ has no consumer (of
    type \f$t\f$), then this last condition is replaced by
    \f$\sigma_{k_u}^t - \sigma_u \geq 1\f$)

    Note that since the matrix of this linear problem is unimodular,
    efficient algorithms (ie polynomial) can be used to solve it.

    A solution of the scheduling problem gives values \f$\sigma_v\f$
    and \f$\sigma_{k_v}^t\f$ for each type \f$t\f$ and each value
    \f$v\f$ of the DDG.

    For each type \f$t\f$, siralina then solves an instance of the
    assignment problem. We now review its definition.

    For each pair of values \f$u, v \in V_R^t\f$, we define a binary
    variable \f$\theta_{u,v}^t \in \{0, 1\}\f$.

    The assignment problem is defined as follows:

    \li minimise \f$\sum\limits_{u,v \in V_R^t} \mu_{u,v}^t
    \theta_{u,v}^t\f$ where \f$\mu_{u,v}^t = \lceil\frac{\sigma_{k_u}^t
    - \delta_w^t(v) - \sigma_v}{p}\rceil\f$

    \li subject to \f$\sum\limits_{v \in V_R^t} \theta_{u,v}^t = 1\f$
    for every \f$u \in V_R^t\f$

    \li and to to \f$\sum\limits_{u \in V_R^t} \theta_{u,v}^t = 1\f$ for
    every \f$v \in V_R^t\f$
    
    The hungarian algorithm can solve efficiently (ie in polynomial
    time) this problem.
    
    A solution of the assignment problem gives values
    \f$\theta_{u,v}^t\f$ that should be interpreted as follows. If
    \f$\theta_{u,v}^t = 1\f$, then \f$(u,v)\f$ becomes a reuse edge
    (of type \f$t\f$ and of reuse distance \f$\mu_{u,v}^t\f$) in the
    reuse graph.

    The number of registers of type \f$t\f$ needed is then
    \f$\sum\limits_{u,v \in V_R^t} \mu_{u,v}^t \theta_{u,v}^t\f$ (ie
    the optimal value found while solving the assigment problem).

    We refer the interested reader to forementioned articles to get
    further details.

    \section solver_api Solver API

    There are three operating modes for the solver: statistics mode,
    search mode and one shot mode.

    \code
    void SIRA_solver_stats(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax);
    \endcode

    This mode is used for statistical analysis. It applies siralina
    solver for every II between IImin and IImax and prints the
    results.

    \code
    int SIRA_solver_search(SIRA_Loop *loop, SIRA_RegConstraintsMap *constraints, SIRA_Param *settings, SIRA_WeightMap *weight_map, int IImin, int IImax, SCEDA_HashMap **solution);
    \endcode

    In this mode, the solver search for the first II (between IImin
    and IImax) that has a solution which satisfies the constraints.

    \code
    int SIRA_solver_single(SIRA_Loop *loop, SIRA_Param *settings, SIRA_WeightMap *weight_map, int II, SCEDA_HashMap **solution);
    \endcode

    In this mode, the siralina solver computes a solution for given II.

    \section solver_exa Example

    \include "solver/main.c"
    
 */
